@@ -38,6 +38,7 @@ import java.security.InvalidKeyException;
 import java.security.MessageDigest;
 import java.security.NoSuchAlgorithmException;
 import java.security.Provider;
+import java.security.spec.AlgorithmParameterSpec;
 import java.util.HashMap;
 import java.util.LinkedHashMap;
 import java.util.List;
@@ -45,12 +46,15 @@ import java.util.Map;
 import static javax.crypto.Cipher.DECRYPT_MODE;
 import static javax.crypto.Cipher.ENCRYPT_MODE;
 import javax.crypto.NoSuchPaddingException;
+import javax.crypto.spec.GCMParameterSpec;
 import javax.crypto.spec.IvParameterSpec;
 import javax.crypto.spec.RC2ParameterSpec;
 
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
+import org.jruby.RubyBoolean;
 import org.jruby.RubyClass;
+import org.jruby.RubyInteger;
 import org.jruby.RubyModule;
 import org.jruby.RubyNumeric;
 import org.jruby.RubyObject;
@@ -557,7 +561,12 @@ public class Cipher extends RubyObject {
             // TODO check cryptoMode CFB/OFB
             final String defaultPadding = "PKCS5Padding";
 
-            if ( padding == null ) return defaultPadding;
+            if ( padding == null ) {
+                if ( "GCM".equalsIgnoreCase(cryptoMode) ) {
+                    return "NoPadding";
+                }
+                return defaultPadding;
+            }
             if ( padding.equalsIgnoreCase("PKCS5Padding") ) {
                 return "PKCS5Padding";
             }
@@ -611,7 +620,9 @@ public class Cipher extends RubyObject {
 
             if ( ivLength == -1 ) {
                 if ( "AES".equals(base) ) {
-                    ivLength = 16;
+                    ivLength = 16; // OpenSSL defaults to 12
+                    // NOTE: we can NOT handle 12 for non GCM mode
+                    if ( "GCM".equals(mode) || "CCM".equals(mode) ) ivLength = 12;
                 }
                 //else if ( "DES".equals(base) ) {
                 //    ivLength = 8;
@@ -790,22 +801,22 @@ public class Cipher extends RubyObject {
     }
 
     @JRubyMethod
-    public IRubyObject name() {
+    public final RubyString name() {
         return getRuntime().newString(name);
     }
 
     @JRubyMethod
-    public IRubyObject key_len() {
+    public final RubyInteger key_len() {
         return getRuntime().newFixnum(keyLength);
     }
 
     @JRubyMethod
-    public IRubyObject iv_len() {
+    public final RubyInteger iv_len() {
         return getRuntime().newFixnum(ivLength);
     }
 
     @JRubyMethod(name = "key_len=", required = 1)
-    public IRubyObject set_key_len(IRubyObject len) {
+    public final IRubyObject set_key_len(IRubyObject len) {
         this.keyLength = RubyNumeric.fix2int(len);
         return len;
     }
@@ -959,7 +970,7 @@ public class Cipher extends RubyObject {
         cipher = getCipherInstance();
     }
 
-    javax.crypto.Cipher getCipherInstance() {
+    final javax.crypto.Cipher getCipherInstance() {
         try {
             return getCipherInstance(realName, false);
         }
@@ -997,7 +1008,11 @@ public class Cipher extends RubyObject {
             throw newCipherError(runtime, "salt must be an 8-octet string");
         }
 
-        final String algorithm = vdigest.isNil() ? "MD5" : ((Digest) vdigest).getAlgorithm();
+        final String algorithm;
+        if ( vdigest.isNil() ) algorithm = "MD5";
+        else {
+            algorithm = (vdigest instanceof Digest) ? ((Digest) vdigest).getAlgorithm() : vdigest.asJavaString();
+        }
         final MessageDigest digest = Digest.getDigest(runtime, algorithm);
         KeyAndIv result = evpBytesToKey(keyLength, ivLength, digest, salt, pass, iter);
         this.key = result.key;
@@ -1040,15 +1055,22 @@ public class Cipher extends RubyObject {
                     );
                 }
                 else {
+                    final AlgorithmParameterSpec ivSpec;
+                    if ( "GCM".equalsIgnoreCase(cryptoMode) ) { // e.g. 'aes-128-gcm'
+                        ivSpec = new GCMParameterSpec(getAuthTagLength() * 8, this.realIV);
+                    }
+                    else {
+                        ivSpec = new IvParameterSpec(this.realIV);
+                    }
                     cipher.init(encryptMode ? ENCRYPT_MODE : DECRYPT_MODE,
                         new SimpleSecretKey(getCipherAlgorithm(), this.key),
-                        new IvParameterSpec(this.realIV)
+                        ivSpec
                     );
                 }
             }
         }
         catch (InvalidKeyException e) {
-            throw newCipherError(runtime, e + ": possibly you need to install Java Cryptography Extension (JCE) Unlimited Strength Jurisdiction Policy Files for your JRE");
+            throw newCipherError(runtime, e + "\n possibly you need to install Java Cryptography Extension (JCE) Unlimited Strength Jurisdiction Policy Files for your JRE");
         }
         catch (Exception e) {
             debugStackTrace(runtime, e);
@@ -1071,29 +1093,35 @@ public class Cipher extends RubyObject {
     @JRubyMethod
     public IRubyObject update(final ThreadContext context, final IRubyObject arg) {
         final Ruby runtime = context.runtime;
+
         if ( isDebug(runtime) ) dumpVars( runtime.getOut(), "update()" );
 
         checkCipherNotNull(runtime);
+        checkAuthTag(runtime);
 
-        final byte[] data = arg.asString().getBytes();
-        if ( data.length == 0 ) {
+        final ByteList data = arg.asString().getByteList();
+        final int length = data.length();
+        if ( length == 0 ) {
             throw runtime.newArgumentError("data must not be empty");
         }
 
-        if ( ! cipherInited ) {
-            //if ( debug ) runtime.getOut().println("BEFORE INITING");
-            doInitCipher(runtime);
-            //if ( debug ) runtime.getOut().println("AFTER INITING");
-        }
+        if ( ! cipherInited ) doInitCipher(runtime);
 
         final ByteList str;
         try {
-            final byte[] out = cipher.update(data);
+            updateAuthData(runtime); // if any
+
+            final byte[] in = data.getUnsafeBytes();
+            final int offset = data.begin();
+            final byte[] out = cipher.update(in, offset, length);
             if ( out != null ) {
                 str = new ByteList(out, false);
-                if ( realIV != null ) setLastIVIfNeeded( encryptMode ? out : data );
+                if ( realIV != null ) {
+                    if ( encryptMode ) setLastIVIfNeeded( out );
+                    else setLastIVIfNeeded( in, offset, length );
+                }
 
-                processedDataBytes += data.length;
+                processedDataBytes += length;
             }
             else {
                 str = new ByteList(ByteList.NULL_ARRAY);
@@ -1115,25 +1143,33 @@ public class Cipher extends RubyObject {
     @JRubyMethod(name = "final")
     public IRubyObject do_final(final ThreadContext context) {
         final Ruby runtime = context.runtime;
+
         checkCipherNotNull(runtime);
+        checkAuthTag(runtime);
+
         if ( ! cipherInited ) doInitCipher(runtime);
         // trying to allow update after final like cruby-openssl. Bad idea.
         if ( "RC4".equalsIgnoreCase(cryptoBase) ) return runtime.newString("");
 
         final ByteList str;
         try {
-            final byte[] out = cipher.doFinal();
-            if ( out != null ) {
-                str = new ByteList(out, false);
-                // TODO: Modifying this line appears to fix the issue, but I do
-                // not have a good reason for why. Best I can tell, lastIv needs
-                // to be set regardless of encryptMode, so we'll go with this
-                // for now. JRUBY-3335.
-                //if ( realIV != null && encryptMode ) ...
-                if ( realIV != null ) setLastIVIfNeeded(out);
+            if ( isAuthDataMode() ) {
+                str = do_final_with_auth(runtime);
             }
             else {
-                str = new ByteList(ByteList.NULL_ARRAY);
+                final byte[] out = cipher.doFinal();
+                if ( out != null ) {
+                    // TODO: Modifying this line appears to fix the issue, but I do
+                    // not have a good reason for why. Best I can tell, lastIv needs
+                    // to be set regardless of encryptMode, so we'll go with this
+                    // for now. JRUBY-3335.
+                    //if ( realIV != null && encryptMode ) ...
+                    str = new ByteList(out, false);
+                    if ( realIV != null ) setLastIVIfNeeded(out);
+                }
+                else {
+                    str = new ByteList(ByteList.NULL_ARRAY);
+                }
             }
 
             //if ( ! isStreamCipher() ) {
@@ -1159,11 +1195,54 @@ public class Cipher extends RubyObject {
         return RubyString.newString(runtime, str);
     }
 
+    private ByteList do_final_with_auth(final Ruby runtime) throws GeneralSecurityException {
+        updateAuthData(runtime); // if any
+
+        final ByteList str;
+        // if GCM/CCM is being used, the authentication tag is appended
+        // in the case of encryption, or verified in the case of decryption.
+        // The result is stored in a new buffer.
+        if ( encryptMode ) {
+            final byte[] out = cipher.doFinal();
+
+            final int len = getAuthTagLength(); int strLen;
+            if ( ( strLen = out.length - len ) > 0 ) {
+                str = new ByteList(out, 0, strLen, false);
+            }
+            else {
+                str = new ByteList(ByteList.NULL_ARRAY); strLen = 0;
+            }
+            auth_tag = new ByteList(out, strLen, out.length - strLen);
+            return str;
+        }
+        else {
+            final byte[] out;
+            if ( auth_tag != null ) {
+                final byte[] tag = auth_tag.getUnsafeBytes();
+                out = cipher.doFinal(tag, auth_tag.begin(), auth_tag.length());
+            }
+            else {
+                out = cipher.doFinal();
+            }
+            return new ByteList(out, false);
+        }
+    }
+
+    private void checkAuthTag(final Ruby runtime) {
+        if ( auth_tag != null && encryptMode ) {
+            throw newCipherError(runtime, "authentication tag already generated by cipher");
+        }
+    }
+
     private void setLastIVIfNeeded(final byte[] tmpIV) {
-        final int len = ivLength;
-        if ( lastIV == null ) lastIV = new byte[len];
-        if ( tmpIV.length >= len ) {
-            System.arraycopy(tmpIV, tmpIV.length - len, lastIV, 0, len);
+        setLastIVIfNeeded(tmpIV, 0, tmpIV.length);
+    }
+
+    private void setLastIVIfNeeded(final byte[] tmpIV, final int offset, final int length) {
+        final int ivLen = this.ivLength;
+        if ( lastIV == null ) lastIV = new byte[ivLen];
+        if ( length >= ivLen ) {
+            System.arraycopy(tmpIV, offset + (length - ivLen), lastIV, 0, ivLen);
         }
     }
 
@@ -1173,6 +1252,70 @@ public class Cipher extends RubyObject {
         return padding;
     }
 
+    private transient ByteList auth_tag;
+
+    @JRubyMethod(name = "auth_tag")
+    public IRubyObject auth_tag(final ThreadContext context) {
+        if ( auth_tag != null ) {
+            return RubyString.newString(context.runtime, auth_tag);
+        }
+        if ( ! isAuthDataMode() ) {
+            throw newCipherError(context.runtime, "authentication tag not supported by this cipher");
+        }
+        return context.nil;
+    }
+
+    @JRubyMethod(name = "auth_tag=")
+    public IRubyObject set_auth_tag(final ThreadContext context, final IRubyObject tag) {
+        if ( ! isAuthDataMode() ) {
+            throw newCipherError(context.runtime, "authentication tag not supported by this cipher");
+        }
+        final RubyString auth_tag = tag.asString();
+        this.auth_tag = StringHelper.setByteListShared(auth_tag);
+        return auth_tag;
+    }
+
+    private boolean isAuthDataMode() { // Authenticated Encryption with Associated Data (AEAD)
+        return "GCM".equalsIgnoreCase(cryptoMode) || "CCM".equalsIgnoreCase(cryptoMode);
+    }
+
+    private static final int MAX_AUTH_TAG_LENGTH = 16;
+
+    private int getAuthTagLength() {
+        return Math.min(MAX_AUTH_TAG_LENGTH, this.key.length); // in bytes
+    }
+
+    private transient ByteList auth_data;
+
+    @JRubyMethod(name = "auth_data=")
+    public IRubyObject set_auth_data(final ThreadContext context, final IRubyObject data) {
+        if ( ! isAuthDataMode() ) {
+            throw newCipherError(context.runtime, "authentication data not supported by this cipher");
+        }
+        final RubyString auth_data = data.asString();
+        this.auth_data = StringHelper.setByteListShared(auth_data);
+        return auth_data;
+    }
+
+    private boolean updateAuthData(final Ruby runtime) {
+        if ( auth_data == null ) return false; // only to be set if auth-mode
+        //try {
+            final byte[] data = auth_data.getUnsafeBytes();
+            cipher.updateAAD(data, auth_data.begin(), auth_data.length());
+        //}
+        //catch (RuntimeException e) {
+        //    debugStackTrace( runtime, e );
+        //    throw newCipherError(runtime, e);
+        //}
+        auth_data = null;
+        return true;
+    }
+
+    @JRubyMethod(name = "authenticated?")
+    public RubyBoolean authenticated_p(final ThreadContext context) {
+        return context.runtime.newBoolean( isAuthDataMode() );
+    }
+
     @JRubyMethod
     public IRubyObject random_key(final ThreadContext context) {
         // str = OpenSSL::Random.random_bytes(self.key_len)
