@@ -24,6 +24,7 @@
 package org.jruby.ext.openssl;
 
 import java.security.NoSuchProviderException;
+import java.security.SecureRandom;
 import java.util.Map;
 
 import org.jruby.CompatVersion;
@@ -35,6 +36,7 @@ import org.jruby.RubyString;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.anno.JRubyModule;
 import org.jruby.runtime.ThreadContext;
+import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.ByteList;
 import org.jruby.util.SafePropertyAccessor;
@@ -56,8 +58,7 @@ public final class OpenSSL {
     }
 
     public static void createOpenSSL(final Ruby runtime) {
-        boolean registerProvider = SafePropertyAccessor.getBoolean("jruby.openssl.provider.register");
-        SecurityHelper.setRegisterProvider( registerProvider );
+        SecurityHelper.setRegisterProvider( SafePropertyAccessor.getBoolean("jruby.openssl.provider.register") );
 
         final RubyModule _OpenSSL = runtime.getOrCreateModule("OpenSSL");
         RubyClass _StandardError = runtime.getClass("StandardError");
@@ -90,16 +91,17 @@ public final class OpenSSL {
         // OpenSSL::VERSION: "1.0.0"
         // OpenSSL::OPENSSL_VERSION: "OpenSSL 1.0.1c 10 May 2012"
         // OpenSSL::OPENSSL_VERSION_NUMBER: 268439615
-        // MRI 1.9.3 / 2.1.2 :
+        // MRI 1.9.3 / 2.2.3 :
         // OpenSSL::VERSION: "1.1.0"
         // OpenSSL::OPENSSL_VERSION: "OpenSSL 1.0.1f 6 Jan 2014"
         // OpenSSL::OPENSSL_VERSION_NUMBER: 268439663
+        // OpenSSL::OPENSSL_LIBRARY_VERSION: ""OpenSSL 1.0.2d 9 Jul 2015"
+        // OpenSSL::FIPS: false
 
         final byte[] version = { '1','.','1','.','0' };
+        final boolean ruby18 = runtime.getInstanceConfig().getCompatVersion() == CompatVersion.RUBY1_8;
+        if ( ruby18 ) version[2] = '0'; // 1.0.0 compatible on 1.8
 
-        if ( runtime.getInstanceConfig().getCompatVersion() == CompatVersion.RUBY1_8 ) {
-            version[2] = '0';
-        } // 1.0.0 compatible on 1.8
         _OpenSSL.setConstant("VERSION", StringHelper.newString(runtime, version));
 
         final RubyModule _Jopenssl = runtime.getModule("Jopenssl");
@@ -107,14 +109,21 @@ public final class OpenSSL {
         final RubyString jVERSION = _Version.getConstantAt("VERSION").asString();
 
         final byte[] JRuby_OpenSSL_ = { 'J','R','u','b','y','-','O','p','e','n','S','S','L',' ' };
-        final int OPENSSL_VERSION_NUMBER = 999999999; // 9469999 do smt useful with it ?
+        final int OPENSSL_VERSION_NUMBER = 999999999; // NOTE: smt more useful?
 
         ByteList OPENSSL_VERSION = new ByteList( jVERSION.getByteList().length() + JRuby_OpenSSL_.length );
         OPENSSL_VERSION.setEncoding( jVERSION.getEncoding() );
         OPENSSL_VERSION.append( JRuby_OpenSSL_ );
         OPENSSL_VERSION.append( jVERSION.getByteList() );
-        _OpenSSL.setConstant("OPENSSL_VERSION", runtime.newString(OPENSSL_VERSION));
+
+        final RubyString VERSION;
+        _OpenSSL.setConstant("OPENSSL_VERSION", VERSION = runtime.newString(OPENSSL_VERSION));
         _OpenSSL.setConstant("OPENSSL_VERSION_NUMBER", runtime.newFixnum(OPENSSL_VERSION_NUMBER));
+        if ( ! ruby18 ) {
+            // MRI 2.3 tests do: /\AOpenSSL +0\./ !~ OpenSSL::OPENSSL_LIBRARY_VERSION
+            _OpenSSL.setConstant("OPENSSL_LIBRARY_VERSION", VERSION);
+            _OpenSSL.setConstant("OPENSSL_FIPS", runtime.getFalse());
+        }
     }
 
     static RubyClass _OpenSSLError(final Ruby runtime) {
@@ -180,6 +189,19 @@ public final class OpenSSL {
 
     // internal (package-level) helpers :
 
+    /**
+     * PRIMARILY MEANT FOR TESTING ONLY, USAGE IS DISCOURAGED!
+     * @see org.jruby.ext.openssl.util.CryptoSecurity
+     */
+    @JRubyMethod(name = "_disable_security_restrictions!", visibility = Visibility.PRIVATE, meta = true)
+    public static IRubyObject _disable_security_restrictions(ThreadContext context, IRubyObject self) {
+        Boolean unrestrict = org.jruby.ext.openssl.util.CryptoSecurity.unrestrictSecurity();
+        Boolean allPerm = org.jruby.ext.openssl.util.CryptoSecurity.setAllPermissionPolicy();
+        if ( unrestrict == null || allPerm == null ) return context.nil;
+        return context.runtime.newBoolean( unrestrict && allPerm );
+    }
+
+
     private static boolean debug;
 
     // on by default, warnings can be disabled using -Djruby.openssl.warn=false
@@ -187,15 +209,15 @@ public final class OpenSSL {
 
     static boolean isDebug() { return debug; }
 
-    static void debugStackTrace(final Throwable e) {
+    public static void debugStackTrace(final Throwable e) {
         if ( isDebug() ) e.printStackTrace(System.out);
     }
 
-    static void debug(final String msg) {
+    public static void debug(final String msg) {
         if ( isDebug() ) System.out.println(msg);
     }
 
-    static void debug(final String msg, final Throwable e) {
+    public static void debug(final String msg, final Throwable e) {
         if ( isDebug() ) System.out.println(msg + ' ' + e);
     }
 
@@ -209,24 +231,31 @@ public final class OpenSSL {
         if ( isDebug(runtime) ) e.printStackTrace(runtime.getOut());
     }
 
-    static void debug(final Ruby runtime, final String msg) {
-        if ( isDebug(runtime) ) runtime.getOut().println(msg);
+    public static void debug(final Ruby runtime, final CharSequence msg) {
+        if ( isDebug(runtime) ) runtime.getOut().println(msg.toString());
     }
 
-    static void debug(final Ruby runtime, final String msg, final Throwable e) {
-        if ( isDebug(runtime) ) runtime.getOut().println(msg + ' ' + e);
+    public static void debug(final Ruby runtime, final CharSequence msg, final Throwable e) {
+        if ( isDebug(runtime) ) runtime.getOut().println(msg.toString() + ' ' + e);
     }
 
-    static void warn(final ThreadContext context, final String msg) {
+    static void warn(final ThreadContext context, final CharSequence msg) {
         warn(context, RubyString.newString(context.runtime, msg));
     }
 
+    static void warn(final ThreadContext context, final RubyString msg) {
+        warn(context, (IRubyObject) msg);
+    }
+
     static void warn(final ThreadContext context, final IRubyObject msg) {
         if ( warn ) context.runtime.getModule("OpenSSL").callMethod(context, "warn", msg);
     }
 
-    private static String javaVersion(final String def) {
-        return SafePropertyAccessor.getProperty("java.version", def);
+    public static String javaVersion(final String def) {
+        final String javaVersionProperty =
+                SafePropertyAccessor.getProperty("java.version", def);
+        if ( javaVersionProperty == "0" ) return "1.7.0"; // Android
+        return javaVersionProperty;
     }
 
     static boolean javaVersion7(final boolean atLeast) {
@@ -239,6 +268,11 @@ public final class OpenSSL {
         return atLeast ? gt <= 0 : gt == 0;
     }
 
+    static boolean javaVersion9(final boolean atLeast) {
+        final int gt = "9".compareTo( javaVersion("0").substring(0, 1) );
+        return atLeast ? gt <= 0 : gt == 0;
+    }
+
     private static String javaName(final String def) {
         // Sun Java 6 or Oracle Java 7/8
         // "Java HotSpot(TM) Server VM" or "Java HotSpot(TM) 64-Bit Server VM"
@@ -247,15 +281,49 @@ public final class OpenSSL {
         return SafePropertyAccessor.getProperty("java.vm.name", def);
     }
 
-    static boolean javaHotSpot() {
+    public static boolean javaHotSpot() {
         return javaName("").contains("HotSpot(TM)");
     }
 
-    static boolean javaOpenJDK() {
+    public static boolean javaOpenJDK() {
         return javaName("").contains("OpenJDK");
     }
 
-    //
+    // shared secure-random :
+
+    private static boolean tryContextSecureRandom = true;
+
+    static SecureRandom getSecureRandom(final Ruby runtime) {
+        return getSecureRandom(runtime, false);
+    }
+
+
+    static SecureRandom getSecureRandom(final Ruby runtime, final boolean nullByDefault) {
+        if ( tryContextSecureRandom ) {
+            SecureRandom random = getSecureRandomFrom(runtime.getCurrentContext());
+            if ( random != null ) return random;
+        }
+        return nullByDefault ? null : new SecureRandom();
+    }
+
+    static SecureRandom getSecureRandomFrom(final ThreadContext context) {
+        if ( tryContextSecureRandom ) {
+            try {
+                SecureRandom random = context.secureRandom;
+                if (random == null) { // public SecureRandom getSecureRandom() on 9K
+                    random = (SecureRandom) context.getClass().getMethod("getSecureRandom").invoke(context);
+                }
+                return random;
+            }
+            catch (Throwable ex) {
+                tryContextSecureRandom = false;
+                debug(context.runtime, "JRuby-OpenSSL failed to retrieve secure random from thread-context", ex);
+            }
+        }
+        return null;
+    }
+
+    // internals
 
     static IRubyObject to_der_if_possible(final ThreadContext context, IRubyObject obj) {
         if ( ! obj.respondsTo("to_der"))  return obj;
