@@ -180,6 +180,7 @@ public class SSLContext extends RubyObject {
         SSLContext.defineAlias("ssl_timeout=", "timeout=");
 
         SSLContext.defineAnnotatedMethods(SSLContext.class);
+        SSLContext.undefineMethod("dup");
 
         final Set<String> methodKeys = SSL_VERSION_OSSL2JSSE.keySet();
         final RubyArray methods = runtime.newArray( methodKeys.size() );
@@ -344,9 +345,7 @@ public class SSLContext extends RubyObject {
         final List<X509AuxCertificate> clientCert;
         if ( value != null && ! value.isNil() ) {
             if ( value.respondsTo("each") ) {
-                final List<X509Cert> cCerts = convertToX509Certs(context, value);
-                clientCert = new ArrayList<X509AuxCertificate>(cCerts.size());
-                for ( X509Cert x : cCerts ) clientCert.add( x.getAuxCert() );
+                clientCert = convertToAuxCerts(context, value);
             } else {
                 if ( ! ( value instanceof X509Cert ) ) {
                     throw runtime.newTypeError("OpenSSL::X509::Certificate expected but got @client_ca = " + value.inspect());
@@ -359,9 +358,7 @@ public class SSLContext extends RubyObject {
         value = getInstanceVariable("@extra_chain_cert");
         final List<X509AuxCertificate> extraChainCert;
         if ( value != null && ! value.isNil() ) {
-            final List<X509Cert> eCerts = convertToX509Certs(context, value);
-            extraChainCert = new ArrayList<X509AuxCertificate>(eCerts.size());
-            for ( X509Cert x : eCerts ) extraChainCert.add( x.getAuxCert() );
+            extraChainCert = convertToAuxCerts(context, value);
         }
         else {
             extraChainCert = null;
@@ -446,9 +443,7 @@ public class SSLContext extends RubyObject {
         */
 
         try {
-            internalContext = new InternalContext(
-                cert, key, store, clientCert, extraChainCert, verifyMode, timeout
-            );
+            internalContext = createInternalContext(context, cert, key, store, clientCert, extraChainCert, verifyMode, timeout);
         }
         catch (GeneralSecurityException e) {
             throw newSSLError(runtime, e);
@@ -469,18 +464,17 @@ public class SSLContext extends RubyObject {
             final Collection<CipherStrings.Def> cipherDefs =
                     CipherStrings.matchingCiphers(this.ciphers, supported, false);
 
-            final RubyArray cipherList = runtime.newArray( cipherDefs.size() );
+            final IRubyObject[] cipherList = new IRubyObject[ cipherDefs.size() ];
 
-            for ( CipherStrings.Def def : cipherDefs ) {
-                final RubyArray cipher = runtime.newArray(4);
-                cipher.store(0, newUTF8String(runtime, def.name));
-                cipher.store(1, newUTF8String(runtime, sslVersionString(def.algorithms)));
-                cipher.store(2, runtime.newFixnum(def.algStrengthBits));
-                cipher.store(3, runtime.newFixnum(def.algBits));
-
-                cipherList.append(cipher);
+            int i = 0; for ( CipherStrings.Def def : cipherDefs ) {
+                cipherList[i++] = runtime.newArrayNoCopy(
+                    newUTF8String(runtime, def.name), // 0
+                    newUTF8String(runtime, sslVersionString(def.algorithms)), // 1
+                    runtime.newFixnum(def.algStrengthBits), // 2
+                    runtime.newFixnum(def.algBits) // 3
+                );
             }
-            return cipherList;
+            return runtime.newArrayNoCopy(cipherList);
         }
         catch (GeneralSecurityException gse) {
             throw newSSLError(runtime, gse.getMessage());
@@ -560,7 +554,7 @@ public class SSLContext extends RubyObject {
         RubyHash DEFAULT_PARAMS = (RubyHash) SSLContext.getConstantAt("DEFAULT_PARAMS");
         if ( args.length == 0 ) params = DEFAULT_PARAMS;
         else {
-            params = (RubyHash) DEFAULT_PARAMS.callMethod(context, "merge", args[0]);
+            params = (RubyHash) DEFAULT_PARAMS.merge(context, args[0], Block.NULL_BLOCK);
         }
         final SSLContext self = this;
         params.visitAll(new RubyHash.Visitor() {
@@ -619,6 +613,12 @@ public class SSLContext extends RubyObject {
         return RubyHash.newHash(context.runtime);
     }
 
+    @JRubyMethod(name = "security_level=")
+    public IRubyObject set_security_level(ThreadContext context, IRubyObject level) {
+        warn(context, "WARNING: SSLContext#security_level= has no effect under JRuby");
+        return context.nil;
+    }
+
     boolean isProtocolForServer() {
         return protocolForServer;
     }
@@ -662,10 +662,12 @@ public class SSLContext extends RubyObject {
     }
 
     // should keep SSLContext as a member for introducin SSLSession. later...
-    SSLEngine createSSLEngine(String peerHost, int peerPort) throws NoSuchAlgorithmException, KeyManagementException {
+    final SSLEngine createSSLEngine(String peerHost, int peerPort)
+        throws NoSuchAlgorithmException, KeyManagementException {
         final SSLEngine engine;
         // an empty peerHost implies no SNI (RFC 3546) support requested
-        if (peerHost == null || peerHost.length() == 0) {
+        if ( peerHost == null || peerHost.length() == 0 ) {
+            // no hints for an internal session reuse strategy
             engine = internalContext.getSSLContext().createSSLEngine();
         }
         // SNI is attempted for valid peerHost hostname on Java >= 7
@@ -795,19 +797,30 @@ public class SSLContext extends RubyObject {
         return 0;
     }
 
-    private List<X509Cert> convertToX509Certs(final ThreadContext context, IRubyObject value) {
-        final ArrayList<X509Cert> result = new ArrayList<X509Cert>();
+    private static List<X509AuxCertificate> convertToAuxCerts(final ThreadContext context, IRubyObject value) {
         final RubyModule SSLContext = _SSLContext(context.runtime);
         final RubyModule Certificate = _Certificate(context.runtime);
+        if ( value instanceof RubyArray ) {
+            final RubyArray val = (RubyArray) value;
+            final int size = val.size();
+            final ArrayList<X509AuxCertificate> result = new ArrayList<X509AuxCertificate>(size);
+            for ( int i=0; i<size; i++ ) result.add( assureCertificate(context, Certificate, val.eltInternal(i)).getAuxCert() );
+            return result;
+        }
+        if ( value instanceof List ) {
+            final List<X509Cert> val = (List) value;
+            final int size = val.size();
+            final ArrayList<X509AuxCertificate> result = new ArrayList<X509AuxCertificate>(size);
+            for ( int i=0; i<size; i++ ) result.add( assureCertificate(context, Certificate, val.get(i)).getAuxCert() );
+            return result;
+        }
+        // else :
+        final ArrayList<X509AuxCertificate> result = new ArrayList<X509AuxCertificate>();
         Utils.invoke(context, value, "each",
             CallBlock.newCallClosure(value, SSLContext, Arity.NO_ARGUMENTS, new BlockCallback() {
 
                 public IRubyObject call(ThreadContext context, IRubyObject[] args, Block block) {
-                    final IRubyObject cert = args[0];
-                    if ( ! ( Certificate.isInstance(cert) ) ) {
-                        throw context.runtime.newTypeError("wrong argument : " + cert.inspect() + " is not a " + Certificate.getName());
-                    }
-                    result.add( (X509Cert) cert );
+                    result.add( assureCertificate(context, Certificate, args[0]).getAuxCert() );
                     return context.nil;
                 }
 
@@ -816,10 +829,26 @@ public class SSLContext extends RubyObject {
         return result;
     }
 
+    private static X509Cert assureCertificate(final ThreadContext context, final RubyModule Certificate, final IRubyObject cert) {
+        if ( ! ( Certificate.isInstance(cert) ) ) {
+            throw context.runtime.newTypeError("wrong argument : " + cert.inspect() + " is not a " + Certificate.getName());
+        }
+        return (X509Cert) cert;
+    }
+
     static RubyClass _SSLContext(final Ruby runtime) {
         return (RubyClass) _SSL(runtime).getConstantAt("SSLContext");
     }
 
+    private InternalContext createInternalContext(ThreadContext context,
+        final X509Cert xCert, final PKey pKey, final Store store,
+        final List<X509AuxCertificate> clientCert, final List<X509AuxCertificate> extraChainCert,
+        final int verifyMode, final int timeout) throws NoSuchAlgorithmException, KeyManagementException {
+        InternalContext internalContext = new InternalContext(xCert, pKey, store, clientCert, extraChainCert, verifyMode, timeout);
+        internalContext.initSSLContext(context);
+        return internalContext;
+    }
+
     /**
      * c: SSL_CTX
      */
@@ -853,8 +882,7 @@ public class SSLContext extends RubyObject {
 
             // initialize SSL context :
 
-            final javax.net.ssl.SSLContext sslContext;
-            sslContext = SecurityHelper.getSSLContext(protocol);
+            final javax.net.ssl.SSLContext sslContext = SecurityHelper.getSSLContext(protocol);
 
             if ( protocolForClient ) {
                 final SSLSessionContext clientContext = sslContext.getClientSessionContext();
@@ -870,11 +898,19 @@ public class SSLContext extends RubyObject {
                     serverContext.setSessionCacheSize(sessionCacheSize);
                 }
             }
+            this.sslContext = sslContext;
+        }
 
+        protected void initSSLContext(final ThreadContext context) throws KeyManagementException {
             final KeyManager[] keyManager = new KeyManager[] { new KeyManagerImpl(this) };
             final TrustManager[] trustManager = new TrustManager[] { new TrustManagerImpl(this) };
-            sslContext.init(keyManager, trustManager, null);
-            this.sslContext = sslContext;
+            // SSLContext (internals) on Sun JDK :
+            // private final java.security.Provider provider; "SunJSSE"
+            // private final javax.net.ssl.SSLContextSpi; sun.security.ssl.SSLContextImpl
+            sslContext.init(keyManager, trustManager, OpenSSL.getSecureRandomFrom(context));
+            // if secureRandom == null JSSE will try :
+            // - new SecureRandom();
+            // - SecureRandom.getInstance("PKCS11", cryptoProvider);
         }
 
         final Store store;
@@ -924,12 +960,8 @@ public class SSLContext extends RubyObject {
 
         @Override
         public String chooseEngineClientAlias(String[] keyType, java.security.Principal[] issuers, javax.net.ssl.SSLEngine engine) {
-            if (internalContext == null) {
-                return null;
-            }
-            if (internalContext.privateKey == null) {
-                return null;
-            }
+            if (internalContext.privateKey == null) return null;
+
             for (int i = 0; i < keyType.length; i++) {
                 if (keyType[i].equalsIgnoreCase(internalContext.keyAlgorithm)) {
                     return keyType[i];
@@ -940,9 +972,8 @@ public class SSLContext extends RubyObject {
 
         @Override
         public String chooseEngineServerAlias(String keyType, java.security.Principal[] issuers, javax.net.ssl.SSLEngine engine) {
-            if (internalContext == null || internalContext.privateKey == null) {
-                return null;
-            }
+            if (internalContext.privateKey == null) return null;
+
             if (keyType.equalsIgnoreCase(internalContext.keyAlgorithm)) {
                 return keyType;
             }
@@ -961,14 +992,14 @@ public class SSLContext extends RubyObject {
 
         @Override // c: ssl3_output_cert_chain
         public java.security.cert.X509Certificate[] getCertificateChain(String alias) {
-            if ( internalContext == null ) return null;
+            final List<java.security.cert.X509Certificate> chain;
 
-            final ArrayList<java.security.cert.X509Certificate> chain =
-                    new ArrayList<java.security.cert.X509Certificate>();
             if ( internalContext.extraChainCert != null ) {
-                chain.addAll(internalContext.extraChainCert);
+                chain = (List) internalContext.extraChainCert;
             }
             else if ( internalContext.cert != null ) {
+                chain = new ArrayList<java.security.cert.X509Certificate>(8);
+
                 StoreContext storeCtx = internalContext.createStoreContext(null);
                 X509AuxCertificate x = internalContext.cert;
                 while (true) {
@@ -995,6 +1026,9 @@ public class SSLContext extends RubyObject {
                     }
                 }
             }
+            else {
+                chain = Collections.EMPTY_LIST;
+            }
             return chain.toArray( new java.security.cert.X509Certificate[chain.size()] );
         }
 
@@ -1005,10 +1039,7 @@ public class SSLContext extends RubyObject {
 
         @Override
         public java.security.PrivateKey getPrivateKey(String alias) {
-            if (internalContext == null || internalContext.privateKey == null) {
-                return null;
-            }
-            return internalContext.privateKey;
+            return internalContext.privateKey; // might be null
         }
 
         @Override
@@ -1039,16 +1070,12 @@ public class SSLContext extends RubyObject {
 
         @Override
         public java.security.cert.X509Certificate[] getAcceptedIssuers() {
-            if ( internalContext == null ) return null;
-
             final int size = internalContext.clientCert.size();
             return internalContext.clientCert.toArray( new java.security.cert.X509Certificate[size] );
         }
 
         // c: ssl_verify_cert_chain
         private void checkTrusted(final String purpose, final X509Certificate[] chain) throws CertificateException {
-            if ( internalContext == null ) throw new CertificateException("uninitialized trust manager");
-
             if ( chain != null && chain.length > 0 ) {
                 if ( (internalContext.verifyMode & SSL.VERIFY_PEER) != 0 ) {
                     // verify_peer
