@@ -32,6 +32,7 @@ import java.net.Socket;
 import java.nio.ByteBuffer;
 import java.nio.channels.Channel;
 import java.nio.channels.ClosedChannelException;
+import java.nio.channels.NotYetConnectedException;
 import java.nio.channels.SelectableChannel;
 import java.nio.channels.SelectionKey;
 import java.nio.channels.Selector;
@@ -40,7 +41,6 @@ import java.security.KeyManagementException;
 import java.security.NoSuchAlgorithmException;
 import java.security.cert.Certificate;
 import java.security.cert.CertificateEncodingException;
-import java.util.Set;
 
 import javax.net.ssl.SSLEngine;
 import javax.net.ssl.SSLEngineResult;
@@ -51,6 +51,7 @@ import javax.net.ssl.SSLPeerUnverifiedException;
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
 import org.jruby.RubyClass;
+import org.jruby.RubyHash;
 import org.jruby.RubyIO;
 import org.jruby.RubyModule;
 import org.jruby.RubyNumeric;
@@ -79,7 +80,7 @@ public class SSLSocket extends RubyObject {
 
     private static final long serialVersionUID = -2084816623554406237L;
 
-    private static ObjectAllocator SSLSOCKET_ALLOCATOR = new ObjectAllocator() {
+    private static final ObjectAllocator ALLOCATOR = new ObjectAllocator() {
         public IRubyObject allocate(Ruby runtime, RubyClass klass) {
             return new SSLSocket(runtime, klass);
         }
@@ -87,13 +88,14 @@ public class SSLSocket extends RubyObject {
 
     public static void createSSLSocket(final Ruby runtime, final RubyModule SSL) { // OpenSSL::SSL
         final ThreadContext context = runtime.getCurrentContext();
-        RubyClass SSLSocket = SSL.defineClassUnder("SSLSocket", runtime.getObject(), SSLSOCKET_ALLOCATOR);
+        RubyClass SSLSocket = SSL.defineClassUnder("SSLSocket", runtime.getObject(), ALLOCATOR);
         // SSLSocket.addReadAttribute(context, "io");
         // SSLSocket.defineAlias("to_io", "io");
         // SSLSocket.addReadAttribute(context, "context");
         SSLSocket.addReadWriteAttribute(context, "sync_close");
         SSLSocket.addReadWriteAttribute(context, "hostname");
         SSLSocket.defineAnnotatedMethods(SSLSocket.class);
+        SSLSocket.undefineMethod("dup");
     }
 
     public SSLSocket(Ruby runtime, RubyClass type) {
@@ -157,6 +159,10 @@ public class SSLSocket extends RubyObject {
             throw runtime.newTypeError("IO expected but got " + args[0].getMetaClass().getName());
         }
         setInstanceVariable("@io", this.io = (RubyIO) args[0]); // compat (we do not read @io)
+        // Ruby 2.3 : @io.nonblock = true if @io.respond_to?(:nonblock=)
+        if (io.respondsTo("nonblock=")) {
+            io.callMethod(context, "nonblock=", runtime.getTrue());
+        }
         setInstanceVariable("@context", this.sslContext); // only compat (we do not use @context)
         // This is a bit of a hack: SSLSocket should share code with
         // RubyBasicSocket, which always sets sync to true.
@@ -186,7 +192,9 @@ public class SSLSocket extends RubyObject {
         peerAppData.limit(0);
         netData.limit(0);
         dummy = ByteBuffer.allocate(0);
-        return this.engine = engine;
+        this.engine = engine;
+        copySessionSetupIfSet();
+        return engine;
     }
 
     @JRubyMethod(name = "io", alias = "to_io")
@@ -207,19 +215,23 @@ public class SSLSocket extends RubyObject {
 
     @JRubyMethod
     public IRubyObject connect(final ThreadContext context) {
-        return connectImpl(context, true);
+        return connectImpl(context, true, true);
     }
 
     @JRubyMethod
-    public IRubyObject connect_nonblock(ThreadContext context) {
-        return connectImpl(context, false);
+    public IRubyObject connect_nonblock(final ThreadContext context) {
+        return connectImpl(context, false, true);
     }
 
-    private SSLSocket connectImpl(final ThreadContext context, final boolean blocking) {
-        final Ruby runtime = context.runtime;
+    @JRubyMethod
+    public IRubyObject connect_nonblock(final ThreadContext context, IRubyObject opts) {
+        return connectImpl(context, false, getExceptionOpt(context, opts));
+    }
+
+    private IRubyObject connectImpl(final ThreadContext context, final boolean blocking, final boolean exception) {
 
         if ( ! sslContext.isProtocolForClient() ) {
-            throw newSSLError(runtime, "called a function you should not call");
+            throw newSSLError(context.runtime, "called a function you should not call");
         }
 
         try {
@@ -230,53 +242,65 @@ public class SSLSocket extends RubyObject {
                 handshakeStatus = engine.getHandshakeStatus();
                 initialHandshake = true;
             }
-            doHandshake(blocking);
+            final IRubyObject ex = doHandshake(blocking, exception);
+            if ( ex != null ) return ex; // :wait_readable | :wait_writable
         }
         catch (SSLHandshakeException e) {
             //debugStackTrace(runtime, e);
             // unlike server side, client should close outbound channel even if
             // we have remaining data to be sent.
             forceClose();
-            throw newSSLErrorFromHandshake(runtime, e);
+            throw newSSLErrorFromHandshake(context.runtime, e);
         }
         catch (NoSuchAlgorithmException e) {
-            debugStackTrace(runtime, e);
+            debugStackTrace(context.runtime, e);
             forceClose();
-            throw newSSLError(runtime, e);
+            throw newSSLError(context.runtime, e);
         }
         catch (KeyManagementException e) {
-            debugStackTrace(runtime, e);
+            debugStackTrace(context.runtime, e);
             forceClose();
-            throw newSSLError(runtime, e);
+            throw newSSLError(context.runtime, e);
         }
         catch (IOException e) {
-            //debugStackTrace(runtime, e);
+            //debugStackTrace(context.runtime, e);
             forceClose();
-            throw newSSLError(runtime, e);
+            throw newSSLError(context.runtime, e);
+        }
+        catch (NotYetConnectedException e) {
+            throw newErrnoEPIPEError(context.runtime, "SSL_connect");
         }
         return this;
     }
 
-    @JRubyMethod
-    public IRubyObject accept(ThreadContext context) {
-        return acceptImpl(context, true);
+    private static RaiseException newErrnoEPIPEError(final Ruby runtime, final String detail) {
+        return Utils.newError(runtime, runtime.getErrno().getClass("EPIPE"), detail);
     }
 
     @JRubyMethod
-    public IRubyObject accept_nonblock(ThreadContext context) {
-        return acceptImpl(context, false);
+    public IRubyObject accept(final ThreadContext context) {
+        return acceptImpl(context, true, true);
+    }
+
+    @JRubyMethod
+    public IRubyObject accept_nonblock(final ThreadContext context) {
+        return acceptImpl(context, false, true);
+    }
+
+    @JRubyMethod
+    public IRubyObject accept_nonblock(final ThreadContext context, IRubyObject opts) {
+        return acceptImpl(context, false, getExceptionOpt(context, opts));
     }
 
     @Deprecated
     public SSLSocket acceptCommon(ThreadContext context, boolean blocking) {
-        return acceptImpl(context, blocking);
+        return (SSLSocket) acceptImpl(context, blocking, true);
     }
 
-    private SSLSocket acceptImpl(final ThreadContext context, final boolean blocking) {
-        final Ruby runtime = context.runtime;
+    private IRubyObject acceptImpl(final ThreadContext context, final boolean blocking, final boolean exception) {
 
         if ( ! sslContext.isProtocolForServer() ) {
-            throw newSSLError(runtime, "called a function you should not call");
+            throw newSSLError(context.runtime, "called a function you should not call");
         }
 
         try {
@@ -301,7 +325,8 @@ public class SSLSocket extends RubyObject {
                 handshakeStatus = engine.getHandshakeStatus();
                 initialHandshake = true;
             }
-            doHandshake(blocking);
+            final IRubyObject ex = doHandshake(blocking, exception);
+            if ( ex != null ) return ex; // :wait_readable | :wait_writable
         }
         catch (SSLHandshakeException e) {
             final String msg = e.getMessage();
@@ -309,34 +334,34 @@ public class SSLSocket extends RubyObject {
             // javax.net.ssl.SSLHandshakeException: No appropriate protocol (protocol is disabled or cipher suites are inappropriate)
             if ( e.getCause() == null && msg != null &&
                  msg.contains("(protocol is disabled or cipher suites are inappropriate)") )  {
-                debug(runtime, sslContext.getProtocol() + " protocol has been deactivated and is not available by default\n see the java.security.Security property jdk.tls.disabledAlgorithms in <JRE_HOME>/lib/security/java.security file");
+                debug(context.runtime, sslContext.getProtocol() + " protocol has been deactivated and is not available by default\n see the java.security.Security property jdk.tls.disabledAlgorithms in <JRE_HOME>/lib/security/java.security file");
             }
             else {
-                debugStackTrace(runtime, e);
+                debugStackTrace(context.runtime, e);
             }
-            throw newSSLErrorFromHandshake(runtime, e);
+            throw newSSLErrorFromHandshake(context.runtime, e);
         }
         catch (NoSuchAlgorithmException e) {
-            debugStackTrace(runtime, e);
-            throw newSSLError(runtime, e);
+            debugStackTrace(context.runtime, e);
+            throw newSSLError(context.runtime, e);
         }
         catch (KeyManagementException e) {
-            debugStackTrace(runtime, e);
-            throw newSSLError(runtime, e);
+            debugStackTrace(context.runtime, e);
+            throw newSSLError(context.runtime, e);
         }
         catch (IOException e) {
-            debugStackTrace(runtime, e);
-            throw newSSLError(runtime, e);
+            debugStackTrace(context.runtime, e);
+            throw newSSLError(context.runtime, e);
         }
         catch (RaiseException e) {
             throw e;
         }
         catch (RuntimeException e) {
-            debugStackTrace(runtime, e);
+            debugStackTrace(context.runtime, e);
             if ( "Could not generate DH keypair".equals( e.getMessage() ) ) {
-                throw SSL.handleCouldNotGenerateDHKeyPairError(runtime, e);
+                throw SSL.handleCouldNotGenerateDHKeyPairError(context.runtime, e);
             }
-            throw newSSLError(runtime, e);
+            throw newSSLError(context.runtime, e);
         }
         return this;
     }
@@ -356,9 +381,10 @@ public class SSLSocket extends RubyObject {
     // SelectableChannel.configureBlocking(false) permanently instead of setting
     // temporarily. SSLSocket requires wrapping IO to be selectable so it should
     // be OK to set configureBlocking(false) permanently.
-    private boolean waitSelect(final int operations, final boolean blocking) throws IOException {
+    private Object waitSelect(final int operations, final boolean blocking, final boolean exception)
+        throws IOException {
         final SocketChannelImpl channel = socketChannelImpl();
-        if ( ! channel.isSelectable() ) return true;
+        if ( ! channel.isSelectable() ) return Boolean.TRUE;
 
         final Ruby runtime = getRuntime();
         final RubyThread thread = runtime.getCurrentContext().getThread();
@@ -368,56 +394,67 @@ public class SSLSocket extends RubyObject {
         final SelectionKey key = channel.register(selector, operations);
 
         try {
-            io.addBlockingThread(thread);
-
             final int[] result = new int[1];
 
-            thread.executeBlockingTask(new RubyThread.BlockingTask() {
-                public void run() throws InterruptedException {
-                    try {
-                        if ( ! blocking ) {
-                            result[0] = selector.selectNow();
+            if ( ! blocking ) {
+                try {
+                    result[0] = selector.selectNow();
 
-                            if ( result[0] == 0 ) {
-                                if ((operations & SelectionKey.OP_READ) != 0 && (operations & SelectionKey.OP_WRITE) != 0) {
-                                    if ( key.isReadable() ) {
-                                        writeWouldBlock(runtime);
-                                    } else if ( key.isWritable() ) {
-                                        readWouldBlock(runtime);
-                                    } else { //neither, pick one
-                                        readWouldBlock(runtime);
-                                    }
-                                } else if ((operations & SelectionKey.OP_READ) != 0) {
-                                    readWouldBlock(runtime);
-                                } else if ((operations & SelectionKey.OP_WRITE) != 0) {
-                                    writeWouldBlock(runtime);
-                                }
+                    if ( result[0] == 0 ) {
+                        if ((operations & SelectionKey.OP_READ) != 0 && (operations & SelectionKey.OP_WRITE) != 0) {
+                            if ( key.isReadable() ) {
+                                writeWouldBlock(runtime, exception, result);
+                            }
+                            //else if ( key.isWritable() ) {
+                            //    readWouldBlock(runtime, exception, result);
+                            //}
+                            else { //neither, pick one
+                                readWouldBlock(runtime, exception, result);
                             }
                         }
-                        else {
-                            result[0] = selector.select();
+                        else if ((operations & SelectionKey.OP_READ) != 0) {
+                            readWouldBlock(runtime, exception, result);
+                        }
+                        else if ((operations & SelectionKey.OP_WRITE) != 0) {
+                            writeWouldBlock(runtime, exception, result);
                         }
                     }
-                    catch (IOException ioe) {
-                        throw runtime.newRuntimeError("Error with selector: " + ioe.getMessage());
+                }
+                catch (IOException ioe) {
+                    throw runtime.newRuntimeError("Error with selector: " + ioe.getMessage());
+                }
+            } else {
+                io.addBlockingThread(thread);
+                thread.executeBlockingTask(new RubyThread.BlockingTask() {
+                    public void run() throws InterruptedException {
+                        try {
+                            result[0] = selector.select();
+                        }
+                        catch (IOException ioe) {
+                            throw runtime.newRuntimeError("Error with selector: " + ioe.getMessage());
+                        }
                     }
-                }
 
-                public void wakeup() {
-                    selector.wakeup();
-                }
-            });
-
-            if ( result[0] >= 1 ) {
-                Set<SelectionKey> keySet = selector.selectedKeys();
-                if ( keySet.iterator().next() == key ) return true;
+                    public void wakeup() {
+                        selector.wakeup();
+                    }
+                });
             }
 
-            return false;
-        }
-        catch (InterruptedException ie) {
-            return false;
+
+            switch ( result[0] ) {
+                case READ_WOULD_BLOCK_RESULT :
+                    return runtime.newSymbol("wait_readable"); // exception: false
+                case WRITE_WOULD_BLOCK_RESULT :
+                    return runtime.newSymbol("wait_writable"); // exception: false
+                case 0 : return Boolean.FALSE;
+                default :
+                    //key should always be contained in selectedKeys() here, however there is a bug in
+                    //JRuby <= 9.1.2.0 that makes this not always the case, so we have to check
+                    return selector.selectedKeys().contains(key) ?  Boolean.TRUE : Boolean.FALSE;
+            }
         }
+        catch (InterruptedException interrupt) { return Boolean.FALSE; }
         finally {
             // Note: I don't like ignoring these exceptions, but it's
             // unclear how likely they are to happen or what damage we
@@ -444,28 +481,41 @@ public class SSLSocket extends RubyObject {
                 debugStackTrace(runtime, e);
             }
 
-            // remove this thread as a blocker against the given IO
-            io.removeBlockingThread(thread);
+            if (blocking) {
+                // remove this thread as a blocker against the given IO
+                io.removeBlockingThread(thread);
 
-            // clear thread state from blocking call
-            thread.afterBlockingCall();
+                // clear thread state from blocking call
+                thread.afterBlockingCall();
+            }
         }
     }
 
-    private static void readWouldBlock(final Ruby runtime) {
-        throw newSSLErrorWaitReadable(runtime, "read would block");
+    private static final int READ_WOULD_BLOCK_RESULT = Integer.MIN_VALUE + 1;
+    private static final int WRITE_WOULD_BLOCK_RESULT = Integer.MIN_VALUE + 2;
+
+    private static void readWouldBlock(final Ruby runtime, final boolean exception, final int[] result) {
+        if ( exception ) throw newSSLErrorWaitReadable(runtime, "read would block");
+        result[0] = READ_WOULD_BLOCK_RESULT;
     }
 
-    private static void writeWouldBlock(final Ruby runtime) {
-        throw newSSLErrorWaitWritable(runtime, "write would block");
+    private static void writeWouldBlock(final Ruby runtime, final boolean exception, final int[] result) {
+        if ( exception ) throw newSSLErrorWaitWritable(runtime, "write would block");
+        result[0] = WRITE_WOULD_BLOCK_RESULT;
     }
 
     private void doHandshake(final boolean blocking) throws IOException {
+        doHandshake(blocking, true);
+    }
+
+    // might return :wait_readable | :wait_writable in case (true, false)
+    private IRubyObject doHandshake(final boolean blocking, final boolean exception) throws IOException {
         while (true) {
-            boolean ready = waitSelect(SelectionKey.OP_READ | SelectionKey.OP_WRITE, blocking);
+            Object sel = waitSelect(SelectionKey.OP_READ | SelectionKey.OP_WRITE, blocking, exception);
+            if ( sel instanceof IRubyObject ) return (IRubyObject) sel; // :wait_readable | :wait_writable
 
             // if not blocking, raise EAGAIN
-            if ( ! blocking && ! ready ) {
+            if ( ! blocking && sel != Boolean.TRUE ) {
                 throw getRuntime().newErrnoEAGAINError("Resource temporarily unavailable");
             }
 
@@ -475,7 +525,7 @@ public class SSLSocket extends RubyObject {
             case FINISHED:
             case NOT_HANDSHAKING:
                 if ( initialHandshake ) finishInitialHandshake();
-                return;
+                return null; // OK
             case NEED_TASK:
                 doTasks();
                 break;
@@ -487,7 +537,8 @@ public class SSLSocket extends RubyObject {
                 // does not mean writable. we explicitly wait for readable channel to avoid
                 // busy loop.
                 if (initialHandshake && status == SSLEngineResult.Status.BUFFER_UNDERFLOW) {
-                    waitSelect(SelectionKey.OP_READ, blocking);
+                    sel = waitSelect(SelectionKey.OP_READ, blocking, exception);
+                    if ( sel instanceof IRubyObject ) return (IRubyObject) sel; // :wait_readable
                 }
                 break;
             case NEED_WRAP:
@@ -669,89 +720,139 @@ public class SSLSocket extends RubyObject {
         flushData(true);
     }
 
-    private RubyString sysreadImpl(final ThreadContext context,
-        final IRubyObject[] args, final boolean blocking) {
+    private IRubyObject sysreadImpl(final ThreadContext context,
+        IRubyObject len, IRubyObject buff, final boolean blocking, final boolean exception) {
         final Ruby runtime = context.runtime;
 
-        final int len = RubyNumeric.fix2int(args[0]);
-        final RubyString buff;
+        final int length = RubyNumeric.fix2int(len);
+        final RubyString buffStr;
 
-        if ( args.length == 2 && ! args[1].isNil() ) {
-            buff = args[1].asString();
+        if ( buff != null && ! buff.isNil() ) {
+            buffStr = buff.asString();
         } else {
-            buff = runtime.newString();
+            buffStr = RubyString.newEmptyString(runtime); // fine since we're setValue
         }
-        if ( len == 0 ) {
-            buff.clear();
-            return buff;
+        if ( length == 0 ) {
+            buffStr.clear();
+            return buffStr;
         }
-        if ( len < 0 ) {
+        if ( length < 0 ) {
             throw runtime.newArgumentError("negative string size (or size too big)");
         }
 
         try {
             // So we need to make sure to only block when there is no data left to process
             if ( engine == null || ! ( peerAppData.hasRemaining() || peerNetData.position() > 0 ) ) {
-                waitSelect(SelectionKey.OP_READ, blocking);
+                final Object ex = waitSelect(SelectionKey.OP_READ, blocking, exception);
+                if ( ex instanceof IRubyObject ) return (IRubyObject) ex; // :wait_readable
             }
 
-            ByteBuffer dst = ByteBuffer.allocate(len);
-            int rr = -1;
+            final ByteBuffer dst = ByteBuffer.allocate(length);
+            int read = -1;
             // ensure >0 bytes read; sysread is blocking read.
-            while ( rr <= 0 ) {
+            while ( read <= 0 ) {
                 if ( engine == null ) {
-                    rr = socketChannelImpl().read(dst);
+                    read = socketChannelImpl().read(dst);
                 } else {
-                    rr = read(dst, blocking);
+                    read = read(dst, blocking);
                 }
 
-                if ( rr == -1 ) throw runtime.newEOFError();
+                if ( read == -1 ) {
+                    if ( exception ) throw runtime.newEOFError();
+                    return runtime.getNil();
+                }
 
-                if ( rr == 0 && status == SSLEngineResult.Status.BUFFER_UNDERFLOW ) {
+                if ( read == 0 && status == SSLEngineResult.Status.BUFFER_UNDERFLOW ) {
                     // If we didn't get any data back because we only read in a partial TLS record,
                     // instead of spinning until the rest comes in, call waitSelect to either block
                     // until the rest is available, or throw a "read would block" error if we are in
                     // non-blocking mode.
-                    waitSelect(SelectionKey.OP_READ, blocking);
+                    final Object ex = waitSelect(SelectionKey.OP_READ, blocking, exception);
+                    if ( ex instanceof IRubyObject ) return (IRubyObject) ex; // :wait_readable
                 }
             }
-            byte[] bss = new byte[rr];
-            dst.position(dst.position() - rr);
-            dst.get(bss);
-            buff.setValue(new ByteList(bss, false));
-            return buff;
+            final byte[] bytesRead = dst.array();
+            final int offset = dst.position() - read;
+            buffStr.setValue(new ByteList(bytesRead, offset, read, false));
+            return buffStr;
         }
         catch (IOException ioe) {
             throw runtime.newIOError(ioe.getMessage());
         }
     }
 
-    @JRubyMethod(rest = true, required = 1, optional = 1)
+    @JRubyMethod
+    public IRubyObject sysread(ThreadContext context, IRubyObject len) {
+        return sysreadImpl(context, len, null, true, true);
+    }
+
+    @JRubyMethod
+    public IRubyObject sysread(ThreadContext context, IRubyObject len, IRubyObject buff) {
+        return sysreadImpl(context, len, buff, true, true);
+    }
+
+    @Deprecated // @JRubyMethod(rest = true, required = 1, optional = 1)
     public IRubyObject sysread(ThreadContext context, IRubyObject[] args) {
-        return sysreadImpl(context, args, true);
+        switch ( args.length) {
+            case 1 :
+                return sysread(context, args[0]);
+            case 2 :
+                return sysread(context, args[0], args[1]);
+        }
+        Arity.checkArgumentCount(context.runtime, args.length, 1, 2);
+        return null; // won't happen as checkArgumentCount raises
     }
 
-    @JRubyMethod(rest = true, required = 1, optional = 2)
+    @JRubyMethod
+    public IRubyObject sysread_nonblock(ThreadContext context, IRubyObject len) {
+        return sysreadImpl(context, len, null, false, true);
+    }
+
+    @JRubyMethod
+    public IRubyObject sysread_nonblock(ThreadContext context, IRubyObject len, IRubyObject arg) {
+        if ( arg instanceof RubyHash ) { // exception: false
+            // NOTE: on Ruby 2.3 this is expected to raise a TypeError (but not on 2.2)
+            return sysreadImpl(context, len, null, false, getExceptionOpt(context, arg));
+        }
+        return sysreadImpl(context, len, arg, false, true); // buffer arg
+    }
+
+    @JRubyMethod
+    public IRubyObject sysread_nonblock(ThreadContext context, IRubyObject len, IRubyObject buff, IRubyObject opts) {
+        return sysreadImpl(context, len, buff, false, getExceptionOpt(context, opts));
+    }
+
+
+    @Deprecated // @JRubyMethod(rest = true, required = 1, optional = 2)
     public IRubyObject sysread_nonblock(ThreadContext context, IRubyObject[] args) {
-        // TODO: options for exception raising
-        return sysreadImpl(context, args, false);
+        switch ( args.length) {
+            case 1 :
+                return sysread_nonblock(context, args[0]);
+            case 2 :
+                return sysread_nonblock(context, args[0], args[1]);
+            case 3 :
+                return sysread_nonblock(context, args[0], args[1], args[2]);
+        }
+        Arity.checkArgumentCount(context.runtime, args.length, 1, 3);
+        return null; // won't happen as checkArgumentCount raises
     }
 
-    private IRubyObject do_syswrite(final ThreadContext context,
-        final IRubyObject arg, final boolean blocking)  {
+    private IRubyObject syswriteImpl(final ThreadContext context,
+        final IRubyObject arg, final boolean blocking, final boolean exception)  {
         final Ruby runtime = context.runtime;
         try {
             checkClosed();
 
-            waitSelect(SelectionKey.OP_WRITE, blocking);
+            final Object ex = waitSelect(SelectionKey.OP_WRITE, blocking, exception);
+            if ( ex instanceof IRubyObject ) return (IRubyObject) ex; // :wait_writable
 
-            ByteList bls = arg.asString().getByteList();
-            ByteBuffer b1 = ByteBuffer.wrap(bls.getUnsafeBytes(), bls.getBegin(), bls.getRealSize());
+            ByteList bytes = arg.asString().getByteList();
+            ByteBuffer buff = ByteBuffer.wrap(bytes.getUnsafeBytes(), bytes.getBegin(), bytes.getRealSize());
             final int written;
             if ( engine == null ) {
-                written = writeToChannel(b1, blocking);
+                written = writeToChannel(buff, blocking);
             } else {
-                written = write(b1, blocking);
+                written = write(buff, blocking);
             }
 
             this.io.callMethod(context, "flush");
@@ -765,18 +866,26 @@ public class SSLSocket extends RubyObject {
 
     @JRubyMethod
     public IRubyObject syswrite(ThreadContext context, IRubyObject arg) {
-        return do_syswrite(context, arg, true);
+        return syswriteImpl(context, arg, true, true);
     }
 
     @JRubyMethod
     public IRubyObject syswrite_nonblock(ThreadContext context, IRubyObject arg) {
-        return do_syswrite(context, arg, false);
+        return syswriteImpl(context, arg, false, true);
     }
 
     @JRubyMethod
-    public IRubyObject syswrite_nonblock(ThreadContext context, IRubyObject arg, IRubyObject options) {
-        // TODO: options for exception raising
-        return do_syswrite(context, arg, false);
+    public IRubyObject syswrite_nonblock(ThreadContext context, IRubyObject arg, IRubyObject opts) {
+        return syswriteImpl(context, arg, false, getExceptionOpt(context, opts));
+    }
+
+    private static boolean getExceptionOpt(final ThreadContext context, final IRubyObject opts) {
+        if ( opts instanceof RubyHash ) { // exception: true
+            final Ruby runtime = context.runtime;
+            IRubyObject exc = ((RubyHash) opts).op_aref(context, runtime.newSymbol("exception"));
+            return exc != runtime.getFalse();
+        }
+        return true;
     }
 
     private void checkClosed() {
@@ -792,7 +901,10 @@ public class SSLSocket extends RubyObject {
     }
 
     private void close(boolean force)  {
-        if ( engine == null ) throw getRuntime().newEOFError();
+        if ( engine == null ) {
+            // if ( force ) throw getRuntime().newEOFError();
+            return;
+        }
 
         engine.closeOutbound();
 
@@ -801,32 +913,38 @@ public class SSLSocket extends RubyObject {
         try {
             doShutdown();
         }
-        catch (IOException e) {
-            // ignore?
+        catch (IOException e) { // ignore?
+            debug(getRuntime(), "SSLSocket.close doShutdown failed", e);
+        }
+        catch (NotYetConnectedException e) {
             debug(getRuntime(), "SSLSocket.close doShutdown failed", e);
         }
     }
 
     @JRubyMethod
     public IRubyObject sysclose(final ThreadContext context) {
+        //if ( isClosed() ) return context.runtime.getNil();
+        if ( this.io.callMethod(context, "closed?").isTrue() ) {
+            return context.runtime.getNil();
+        } // Ruby 2.3
         // no need to try shutdown when it's a server
         close( sslContext.isProtocolForClient() );
 
         if ( this.callMethod(context, "sync_close").isTrue() ) {
-            this.io.callMethod(context, "close");
+            return this.io.callMethod(context, "close");
         }
         return context.runtime.getNil();
     }
 
     @JRubyMethod
-    public IRubyObject cert() {
-        final Ruby runtime = getRuntime();
+    public IRubyObject cert(final ThreadContext context) {
+        final Ruby runtime = context.runtime;
         if ( engine == null ) return runtime.getNil();
 
         try {
             Certificate[] cert = engine.getSession().getLocalCertificates();
             if ( cert != null && cert.length > 0 ) {
-                return X509Cert.wrap(runtime, cert[0]);
+                return X509Cert.wrap(context, cert[0]);
             }
         }
         catch (CertificateEncodingException e) {
@@ -835,40 +953,50 @@ public class SSLSocket extends RubyObject {
         return runtime.getNil();
     }
 
+    // @Deprecated
+    public final IRubyObject cert() {
+        return cert(getRuntime().getCurrentContext());
+    }
+
     @JRubyMethod
-    public IRubyObject peer_cert() {
-        final Ruby runtime = getRuntime();
+    public IRubyObject peer_cert(final ThreadContext context) {
+        final Ruby runtime = context.runtime;
         if ( engine == null ) return runtime.getNil();
 
         try {
             Certificate[] cert = engine.getSession().getPeerCertificates();
             if ( cert.length > 0 ) {
-                return X509Cert.wrap(runtime, cert[0]);
+                return X509Cert.wrap(context, cert[0]);
             }
         }
         catch (CertificateEncodingException e) {
             throw X509Cert.newCertificateError(runtime, e);
         }
         catch (SSLPeerUnverifiedException e) {
-            if (runtime.isVerbose() || OpenSSL.isDebug(runtime)) {
+            if (OpenSSL.isDebug(runtime)) {
                 runtime.getWarnings().warning(String.format("%s: %s", e.getClass().getName(), e.getMessage()));
             }
         }
         return runtime.getNil();
     }
 
+    // @Deprecated
+    public final IRubyObject peer_cert() {
+        return peer_cert(getRuntime().getCurrentContext());
+    }
+
     @JRubyMethod
-    public IRubyObject peer_cert_chain() {
-        final Ruby runtime = getRuntime();
+    public IRubyObject peer_cert_chain(final ThreadContext context) {
+        final Ruby runtime = context.runtime;
         if ( engine == null ) return runtime.getNil();
 
         try {
             javax.security.cert.Certificate[] certs = engine.getSession().getPeerCertificateChain();
-            RubyArray arr = runtime.newArray(certs.length);
+            IRubyObject[] cert_chain = new IRubyObject[ certs.length ];
             for ( int i = 0; i < certs.length; i++ ) {
-                arr.append( X509Cert.wrap(runtime, certs[i]) );
+                cert_chain[i] = X509Cert.wrap(context, certs[i]);
             }
-            return arr;
+            return runtime.newArrayNoCopy(cert_chain);
         }
         catch (javax.security.cert.CertificateEncodingException e) {
             throw X509Cert.newCertificateError(getRuntime(), e);
@@ -881,12 +1009,25 @@ public class SSLSocket extends RubyObject {
         return runtime.getNil();
     }
 
+    // @Deprecated
+    public final IRubyObject peer_cert_chain() {
+        return peer_cert_chain(getRuntime().getCurrentContext());
+    }
+
     @JRubyMethod
     public IRubyObject cipher() {
         if ( engine == null ) return getRuntime().getNil();
         return getRuntime().newString( engine.getSession().getCipherSuite() );
     }
 
+    @JRubyMethod
+    public IRubyObject npn_protocol() {
+        if ( engine == null ) return getRuntime().getNil();
+        // NOTE: maybe a time to use https://github.com/benmmurphy/ssl_npn
+        warn(getRuntime().getCurrentContext(), "WARNING: SSLSocket#npn_protocol is not supported");
+        return getRuntime().getNil(); // throw new UnsupportedOperationException();
+    }
+
     @JRubyMethod
     public IRubyObject state() {
         warn(getRuntime().getCurrentContext(), "WARNING: unimplemented method called: SSLSocket#state");
@@ -899,13 +1040,33 @@ public class SSLSocket extends RubyObject {
         return getRuntime().getNil();
     }
 
-    @JRubyMethod(name = "session_reused?")
-    public IRubyObject session_reused_p() {
-        warn(getRuntime().getCurrentContext(), "WARNING: SSLSocket#session_reused? is not supported");
-        return getRuntime().getNil(); // throw new UnsupportedOperationException();
+    private boolean reusableSSLEngine() {
+        if ( engine != null ) {
+            final String peerHost = engine.getPeerHost();
+            if ( peerHost != null && peerHost.length() > 0 ) {
+                // NOT getSSLContext().createSSLEngine() - no hints for session reuse
+                return true;
+            }
+        }
+        return false;
     }
 
-    javax.net.ssl.SSLSession getSession() {
+    @JRubyMethod(name = "session_reused?")
+    public IRubyObject session_reused_p() {
+        if ( reusableSSLEngine() ) {
+            if ( ! engine.getEnableSessionCreation() ) {
+                // if session creation is disabled we can be sure its to be re-used
+                return getRuntime().getTrue();
+            }
+            //return getRuntime().getFalse(); // NOTE: likely incorrect (we can not decide)
+        }
+        //warn(getRuntime().getCurrentContext(), "WARNING: SSLSocket#session_reused? is not supported");
+        return getRuntime().getNil(); // can not decide - probably not
+    }
+
+    // JSSE: SSL Sessions can be reused only if connecting to the same host at the same port
+
+    final javax.net.ssl.SSLSession sslSession() {
         return engine == null ? null : engine.getSession();
     }
 
@@ -913,17 +1074,41 @@ public class SSLSocket extends RubyObject {
 
     @JRubyMethod(name = "session")
     public IRubyObject session(final ThreadContext context) {
-        if ( getSession() == null ) return context.nil;
+        if ( sslSession() == null ) return context.nil;
+        return getSession(context.runtime);
+    }
+
+    private SSLSession getSession(final Ruby runtime) {
         if ( session == null ) {
-            return session = new SSLSession(context.runtime).initializeImpl(context, this);
+            return session = new SSLSession(runtime).initializeImpl(this);
         }
         return session;
     }
 
+    private transient SSLSession setSession = null;
+
     @JRubyMethod(name = "session=")
     public IRubyObject set_session(IRubyObject session) {
-        warn(getRuntime().getCurrentContext(), "WARNING: SSLSocket#session= is not supported");
-        return getRuntime().getNil(); // throw new UnsupportedOperationException();
+        // NOTE: we can not fully support this without the SSL provider internals
+        // but we can assume setting a session= is meant as a forced session re-use
+        if ( session instanceof SSLSession ) {
+            setSession = (SSLSession) session;
+            if ( engine != null ) copySessionSetupIfSet();
+        }
+        //warn(context, "WARNING: SSLSocket#session= has not effect");
+        return getRuntime().getNil();
+    }
+
+    private void copySessionSetupIfSet() {
+        if ( setSession != null ) {
+            if ( reusableSSLEngine() ) {
+                engine.setEnableSessionCreation(false);
+                final ThreadContext context = getRuntime().getCurrentContext();
+                if ( ! setSession.equals( getSession(context.runtime) ) ) {
+                    getSession(context.runtime).set_timeout(context, setSession.timeout(context));
+                }
+            }
+        }
     }
 
     @JRubyMethod
@@ -947,7 +1132,7 @@ public class SSLSocket extends RubyObject {
         throw new IllegalStateException("unknow channel impl: " + channel + " of type " + channel.getClass().getName());
     }
 
-    private static interface SocketChannelImpl {
+    private interface SocketChannelImpl {
 
         boolean isOpen() ;
 
