@@ -70,6 +70,11 @@ final class Utils {
         return new RaiseException(runtime, runtime.getRuntimeError(), msg, true);
     }
 
+    static RaiseException newErrorWithoutTrace(Ruby runtime, RubyClass errorClass, String message, boolean nativeException) {
+        final IRubyObject backtrace = runtime.newEmptyArray(); // runtime.getNil();
+        return new RaiseException(runtime, errorClass, message, backtrace, nativeException);
+    }
+
     static RaiseException newError(Ruby runtime, RubyClass errorClass, String message, boolean nativeException) {
         return new RaiseException(runtime, errorClass, message, nativeException);
     }
