@@ -73,6 +73,7 @@ public class X509StoreContext extends RubyObject {
     public static void createX509StoreContext(final Ruby runtime, final RubyModule X509) {
         RubyClass StoreContext = X509.defineClassUnder("StoreContext", runtime.getObject(), X509STORECTX_ALLOCATOR);
         StoreContext.defineAnnotatedMethods(X509StoreContext.class);
+        StoreContext.undefineMethod("dup");
     }
 
     private static RubyClass _StoreContext(final Ruby runtime) {
@@ -114,13 +115,30 @@ public class X509StoreContext extends RubyObject {
             if ( args.length > 2) chain = args[2];
         }
 
-        final X509AuxCertificate _cert = cert.isNil() ? null : ((X509Cert) cert).getAuxCert();
+        final X509AuxCertificate _cert;
+        if (cert.isNil()) {
+            _cert = null;
+        }
+        else {
+            if ( ! (cert instanceof X509Cert) ) {
+                throw context.runtime.newTypeError(cert, "OpenSSL::X509::Certificate");
+            }
+            _cert = ((X509Cert) cert).getAuxCert();
+        }
         final List<X509AuxCertificate> _chain;
         if ( ! chain.isNil() ) {
             @SuppressWarnings("unchecked")
-            final List<X509Cert> certs = (List<X509Cert>) chain; // RubyArray
+            final RubyArray certs = (RubyArray) chain;
             _chain = new ArrayList<X509AuxCertificate>( certs.size() );
-            for ( X509Cert x : certs ) _chain.add( x.getAuxCert() );
+
+            for (int i = 0; i < certs.size(); i++) {
+                // NOTE: if we use the normal java syntax for iterating over this
+                // RubyArray, the `toJava` method of the X509Cert class will be
+                // implicitly called, and that will return the BC certificate object
+                // rather than the JRuby one.
+                X509Cert c = (X509Cert) certs.eltOk(i);
+                _chain.add(c.getAuxCert());
+            }
         }
         else {
             _chain = new ArrayList<X509AuxCertificate>(4);
