@@ -27,6 +27,7 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby.ext.openssl.x509store;
 
+import org.jruby.ext.openssl.OpenSSL;
 import org.jruby.ext.openssl.util.Cache;
 import static org.jruby.ext.openssl.x509store.X509Utils.X509_CERT_DIR;
 import static org.jruby.ext.openssl.x509store.X509Utils.X509_FILETYPE_ASN1;
@@ -352,13 +353,14 @@ public class Lookup {
         }
     }
 
-    public int loadDefaultJavaCACertsFile() throws IOException, GeneralSecurityException {
-        final String certsFile = X509Utils.X509_CERT_FILE.replace('/', File.separatorChar);
+    public int loadDefaultJavaCACertsFile(String certsFile) throws IOException, GeneralSecurityException {
         final FileInputStream fin = new FileInputStream(certsFile);
         int count = 0;
         try {
-            KeyStore keystore = SecurityHelper.getKeyStore(KeyStore.getDefaultType());
-            // we pass a null password, as the cacerts file isn't password protected
+	        // hardcode the keystore type, as we expect cacerts to be a java
+	        // keystore - especially needed for jdk9
+            KeyStore keystore = SecurityHelper.getKeyStore("jks");
+	        // null password - as the cacerts file isn't password protected
             keystore.load(fin, null);
             PKIXParameters params = new PKIXParameters(keystore);
             for ( TrustAnchor trustAnchor : params.getTrustAnchors() ) {
@@ -514,18 +516,20 @@ public class Lookup {
             case X509_L_FILE_LOAD:
                 if (arglInt == X509_FILETYPE_DEFAULT) {
                     try {
-                        file = ctx.envEntry( getDefaultCertificateFileEnvironment() );
+                        file = ctx.envEntry( getDefaultCertificateFileEnvironment() ); // ENV['SSL_CERT_FILE']
                     }
-                    catch (RuntimeException e) { }
-
-                    if (file != null) {
+                    catch (RuntimeException e) {
+                        OpenSSL.debug(ctx.runtime, "failed to read SSL_CERT_FILE", e);
+                    }
+                    if (file == null) {
+                        file = X509Utils.X509_CERT_FILE.replace('/', File.separatorChar);
+                    }
+                    if (file.matches(".*\\.(crt|cer|pem)$")) {
                         ok = ctx.loadCertificateOrCRLFile(file, X509_FILETYPE_PEM) != 0 ? 1 : 0;
                     } else {
-                        ok = (ctx.loadDefaultJavaCACertsFile() != 0) ? 1: 0;
-                    }
-                    if (ok == 0) {
-                        X509Error.addError(X509_R_LOADING_DEFAULTS);
+                        ok = (ctx.loadDefaultJavaCACertsFile(file) != 0) ? 1: 0;
                     }
+                    // we ignore errors on loading default paths the same way MRI does it
                 } else {
                     if (arglInt == X509_FILETYPE_PEM) {
                         ok = (ctx.loadCertificateOrCRLFile(argp, X509_FILETYPE_PEM) != 0) ? 1 : 0;
